<?php

namespace Drupal\contribution_records\Drush\Commands;

use Drupal\Component\Utility\UrlHelper;
use Drupal\contribution_records\SourceLink\DrupalOrgIssue;
use Drupal\Core\Cache\MemoryCache\MemoryCache;
use Drupal\Core\Database\Database;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Queue\QueueFactory;
use Drupal\node\Entity\Node;
use Drush\Attributes as CLI;
use Drush\Commands\AutowireTrait;
use Drush\Commands\DrushCommands as BaseDrushCommands;
use Symfony\Component\DependencyInjection\Attribute\Autowire;

/**
 * A drush command file.
 *
 * @package Drupal\drupalorg_migrate\Commands
 */
final class DrushCommands extends BaseDrushCommands {

  use AutowireTrait;

  /**
   * Size of the batch to use.
   *
   * @var int
   */
  const BATCH_LENGTH = 50;

  /**
   * Max number of issues found not to be migrated before exiting.
   *
   * @var int
   */
  const LIMIT_FOUND = 500;

  /**
   * {@inheritdoc}
   */
  public function __construct(
    #[Autowire(service: 'entity_type.manager')]
    protected EntityTypeManagerInterface $entityTypeManager,
    #[Autowire(service: 'entity.memory_cache')]
    protected MemoryCache $memoryCache,
    #[Autowire(service: 'queue')]
    protected QueueFactory $queueFactory,
  ) {
    parent::__construct();
  }

  /**
   * Send the given issues from D7 to the import queue.
   */
  #[CLI\Command(name: 'contribution_records:sync-contribution-records')]
  #[CLI\Option(name: 'nids', description: 'Issue node IDs on D7.')]
  #[CLI\Option(name: 'hard-reset', description: 'Hard reset the Contribution Record contributors.')]
  #[CLI\Option(name: 'base-url', description: 'Domain where the nodes come from (eg: https://www.drupal.org).')]
  public function syncContributionRecords(
    $options = [
      'nids' => NULL,
      'hard-reset' => 0,
      'base-url' => 'https://www.drupal.org',
    ],
  ) {
    $nids = [];
    $base_url = $options['base-url'];
    $hard_reset = (bool) $options['hard-reset'];
    if (!empty($options['nids'])) {
      $nids = explode(',', $options['nids']);
    }
    if (!UrlHelper::isValid($base_url, TRUE)) {
      if (!empty($base_url)) {
        throw new \Exception('Domain provided is not a valid URL.');
      }
      $base_url = NULL;
    }

    if (!is_null($base_url) && !empty($nids)) {
      foreach ($nids as $nid) {
        $url = $base_url . '/node/' . $nid;
        $this->queueFactory->get('contribution_records_import_queue_worker')->createItem([
          'url' => $url,
          'sync' => TRUE,
          'hard-reset' => $hard_reset,
        ]);
        $this->io()->info(\dt('Queued @url.', [
          '@url' => $url,
        ]));
      }
    }
  }

  /**
   * Check contribution records in D7 and not in D10.
   */
  #[CLI\Command(name: 'contribution_records:check-import-status')]
  #[CLI\Option(name: 'types', description: 'Types of nodes to check (eg: project_issue, sa).')]
  #[CLI\Option(name: 'base-url', description: 'Domain where the nodes come from (eg: https://www.drupal.org).')]
  public function checkEntities($options = ['types' => NULL, 'base-url' => 'https://www.drupal.org']) {
    $types = DrupalOrgIssue::ALLOWED_NODE_TYPES;
    if (!empty($options['types'])) {
      $types = explode(',', $options['types']);
      if (!empty(array_diff($types, DrupalOrgIssue::ALLOWED_NODE_TYPES))) {
        throw new \Exception('Invalid type(s). Allowed values: ' . implode(', ', DrupalOrgIssue::ALLOWED_NODE_TYPES));
      }
    }
    if (!UrlHelper::isValid($options['base-url'], TRUE)) {
      if (!empty($options['base-url'])) {
        throw new \Exception('Domain provided is not a valid URL.');
      }
      $options['base-url'] = NULL;
    }

    $this->io()->info(\dt('Checking Contribution Records import status.'));
    foreach ($types as $type) {
      $this->io()->info(\dt('Checking @type nodes.', [
        '@type' => $type,
      ]));
      $not_imported = $this->contributionRecordsNotImported($type, $options['base-url']);
      if (empty($not_imported)) {
        $this->io()->success(\dt('All Contribution Records of type "@type" are in the new system.', [
          '@type' => $type,
        ]));
      }
      else {
        $count_not_imported = count($not_imported);
        $issue_ids_as_string = $this->arrayToStringTruncated($not_imported);
        $this->io()->warning(\dt('@count Contribution Records of type "@type" are not migrated, with the following Drupal 7 issue IDs: @ids', [
          '@count' => $count_not_imported,
          '@type' => $type,
          '@ids' => $issue_ids_as_string,
        ]));
      }
    }
  }

  /**
   * Show array as string (comma separated) and truncate it if too long.
   *
   * @param array $array
   *   Array to convert to string.
   *
   * @return string
   *   String representation of the array.
   */
  protected function arrayToStringTruncated($array) {
    return (count($array) > 100) ?
      implode(', ', array_slice($array, 0, 100)) . '... (truncated)' :
      implode(', ', $array);
  }

  /**
   * Find contribution records not imported.
   *
   * @param string $type
   *   Type of node in the source database.
   * @param string $base_url
   *   (Optional) Base URL where the records come from.
   *
   * @return array
   *   List of Drupal issue ids.
   */
  protected function contributionRecordsNotImported($type, $base_url = NULL) {
    $count_entities = $this->getIssueCount($type);
    $length = self::BATCH_LENGTH;
    $offset = 0;
    $extra_entities = [];
    $verbose = $this->output()->isVerbose();

    $ids = $this->getNextIssuesBatch($type, $offset, $length);
    $this->io()->progressStart($count_entities);
    while (
      !empty($ids) &&
      (count($extra_entities) < self::LIMIT_FOUND || $verbose)
    ) {
      $ids_not_in_database = $this->issuesNotImportedAsContributionRecords($ids, $base_url);
      if ($verbose && !empty($ids_not_in_database)) {
        $this->io()->text("\nBatch of '$type' IDs not found: " . implode(',', $ids_not_in_database));
      }
      $extra_entities = array_merge($extra_entities, $ids_not_in_database);

      $this->io()->progressAdvance($length);

      $offset += $length;
      $ids = $this->getNextIssuesBatch($type, $offset, $length);
    }
    $this->io()->progressFinish();

    return $extra_entities;
  }

  /**
   * Gets a count of issues.
   *
   * @param string $type
   *   Type of node.
   *
   * @return int
   *   Count of entities.
   */
  protected function getIssueCount($type) {
    $connection = Database::getConnection('default', 'migrate');
    $query = $connection
      ->select('node', 'e')
      ->condition('e.type', $type);
    return $query
      ->countQuery()
      ->execute()
      ->fetchField();
  }

  /**
   * Get a batch of entity ids to check.
   *
   * @param string $type
   *   Type of nodes.
   * @param int $offset
   *   Offset of the records to get.
   * @param int $length
   *   Length of the records to get.
   *
   * @return array
   *   Array of entity ids.
   */
  protected function getNextIssuesBatch($type, $offset, $length) {
    $connection = Database::getConnection('default', 'migrate');
    $query = $connection
      ->select('node', 'e')
      ->fields('e', ['nid'])
      ->condition('e.type', $type)
      ->condition('e.status', 1)
      ->range($offset, $length)
      ->orderBy("e.nid");
    return $query
      ->execute()
      ->fetchCol();
  }

  /**
   * Check which entity ids are not in a database.
   *
   * @param array $ids
   *   List of entity ids to check.
   * @param string $base_url
   *   (Optional) Base URL where the records come from.
   *
   * @return array
   *   Array of entity ids not present in the given database.
   */
  protected function issuesNotImportedAsContributionRecords($ids, ?string $base_url = NULL) {
    $query = $this->entityTypeManager->getStorage('node')->getQuery();
    $orGroup = $query->orConditionGroup();
    $links = [];
    foreach ($ids as $id) {
      // Contribution records from Drupal.org use the canonical /node/NID URL.
      if (!is_null($base_url)) {
        $orGroup->condition('field_source_link', $base_url . '/node/' . $id);
      }
      else {
        $orGroup->condition('field_source_link', '%/node/' . $id, 'LIKE');
      }
      $links[$id] = '/node/' . $id;
    }

    $matched_ids = $query
      ->accessCheck(FALSE)
      ->condition('type', 'contribution_record')
      ->condition($orGroup)
      ->execute();

    $found_ids = [];
    if (count($matched_ids) === count($ids)) {
      $found_ids = $ids;
    }
    else {
      $nodes = Node::loadMultiple($matched_ids);
      foreach ($nodes as $node) {
        $link = $node->get('field_source_link')->getValue()[0]['uri'] ?? NULL;
        if (str_contains($link, '/node/')) {
          $link = strstr($link, '/node/');
          $found_key = array_search($link, $links);
          if ($found_key !== FALSE) {
            $found_ids[] = $found_key;
          }
        }
      }
      $this->memoryCache->deleteAll();
    }

    return array_diff($ids, $found_ids);
  }

}
