<?php

namespace Drupal\contribution_records\Plugin\QueueWorker;

use Drupal\contribution_records\SourceLink;
use Drupal\Core\DependencyInjection\ClassResolverInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Queue\QueueWorkerBase;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Defines 'contribution_records_import_queue_worker' queue worker.
 *
 * Run via `drush` like this:
 * `drush queue:run contribution_records_import_queue_worker`.
 *
 * @QueueWorker(
 *   id = "contribution_records_import_queue_worker",
 *   title = @Translation("Import Queue Worker")
 * )
 */
class ContributionRecordImportQueueWorker extends QueueWorkerBase implements ContainerFactoryPluginInterface {

  use StringTranslationTrait;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('logger.factory')->get('contribution_records'),
      $container->get('class_resolver')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    protected LoggerInterface $logger,
    protected ClassResolverInterface $classResolver,
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
  }

  /**
   * {@inheritdoc}
   */
  public function processItem($data) {
    if (!$this->validateItem($data)) {
      return FALSE;
    }

    /** @var \Drupal\contribution_records\SourceLink $source_link_instance */
    $source_link_instance = $this->classResolver->getInstanceFromDefinition(SourceLink::class);
    $source_link = $source_link_instance->setLink($data['url']);
    if ($source_link->isValid(FALSE)) {
      $entity = $source_link->isDuplicated(TRUE);
      $sync = $data['sync'] ?? FALSE;
      // Trigger a re-sync from the source. Ignore the rest of the data.
      if ($sync && $entity) {
        $reset_contributors = $data['hard-reset'] ?? FALSE;
        $source_link->syncContribRecord($entity, TRUE, $reset_contributors);
        $this->logger->info($this->t('Syncing @url.', [
          '@url' => $data['url'],
        ]));
      }
      // Data with the 'sync' does not have all the raw information.
      elseif (!$sync) {
        if (!$entity) {
          $source_link->createContribRecordFromRawData($data);
        }
        else {
          $this->logger->notice($this->t('The item with URL @url is already imported.', [
            '@url' => $data['url'],
          ]));
        }
      }
      else {
        $this->logger->warning($this->t('The item with URL @url is not in the system yet.', [
          '@url' => $data['url'],
        ]));
      }
    }
    else {
      $this->logger->error($this->t('The item with URL @url cannot be imported due to: @error', [
        '@url' => $data['url'],
        '@error' => $source_link->getError(),
      ]));
    }
  }

  /**
   * Validate item array to make sure all key elements are there.
   *
   * @param array $data
   *   Item to validate.
   *
   * @return bool
   *   Whether the item was valid or not.
   */
  protected function validateItem(array $data) {
    // All we need for a re-sync is the URL.
    if (!empty($data['sync']) && !empty($data['url'])) {
      return TRUE;
    }

    // Otherwise we need at least URL, title and credits.
    if (
      empty($data['url']) ||
      empty($data['title']) ||
      empty($data['credits'])
    ) {
      $this->logger->error($this->t('Not all required fields are present. URL: @url // Title: @title // Credits count: @count', [
        '@url' => $data['url'] ?? '-',
        '@title' => $data['title'] ?? '-',
        '@count' => !empty($data['credits']) ? count($data['credits']) : 0,
      ]));
      return FALSE;
    }

    return TRUE;
  }

}
